@@ -11,9 +11,12 @@ from .logic.bitwise import *
 from .logic.storage import *
 from .schematic_symbols import *
 
-def getVerilogModuleName(obj:Logic):
-    sid = hex(id(obj))
-    str = type(obj).__name__ + "_" + sid[2:] 
+def getVerilogModuleName(obj:Logic, noInstanceNumber=False):
+    str = type(obj).__name__  
+    if (not(noInstanceNumber)):
+        sid = hex(id(obj))
+        str += "_" +sid[2:] 
+        
     return str
 
 def getWidthInfo(w:Wire):
@@ -106,7 +109,7 @@ def getWireNames(obj:Logic):
     ret = {}
     
     # process all the wires connected to child instances
-    for child in obj.children:
+    for child in obj.children.values():
         for wire in collectPortWires(child):
             ret[wire]= "w_" + wire.name
  
@@ -137,10 +140,12 @@ def collectPortWires( obj:Logic):
     ret = []
     
     for inp in obj.inPorts:
-        ret.append(inp.wire)
+        if (not(inp.wire is None)):
+            ret.append(inp.wire)
 
     for outp in obj.outPorts:
-        ret.append(outp.wire)
+        if (not(outp.wire is None)):
+            ret.append(outp.wire)
 
     return ret
 
@@ -163,7 +168,7 @@ def collectLocalWires(obj:Logic):
     
     portWires = collectPortWires(obj)
     
-    for child in obj.children:
+    for child in obj.children.values():
         
         childLocalWires = [x for x in collectPortWires(child) if x not in portWires]
         ret.extend(childLocalWires)
@@ -209,6 +214,9 @@ def InlineNand2(obj:Logic):
 def InlineOr2(obj:Logic):
     return "assign {} = {} | {};\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.a) , getParentWireName(obj, obj.b))
 
+def InlineNor2(obj:Logic):
+    return "assign {} = ~({} | {});\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.a) , getParentWireName(obj, obj.b))
+
 def InlineXor2(obj:Logic):
     return "assign {} = {} ^ {};\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.a) , getParentWireName(obj, obj.b))
 
@@ -224,13 +232,66 @@ def InlineSub(obj:Logic):
 def InlineEqualConstant(obj:Logic):
     return "assign {} = ({} == {})? 1 : 0;\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.a), obj.v )
 
-def InlineBits(obj:Logic):
+def InlineRange(obj:Logic):
+    return "assign {} = {}[{}:{}];\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.a) , obj.high, obj.low)
+
+def InlineBit(obj:Logic):
+    return "assign {} = {}[{}];\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.a) , obj.bit)
+
+def InlineBitsLSBF(obj:Logic):
     str = ""
-    for i in range(len(obj.bits)):
+    w = len(obj.bits)
+    if (w == 1):
+        return "assign {} = {};\n".format(getParentWireName(obj, obj.bits[0]), getParentWireName(obj, obj.a))
+
+    for i in range(w):
         str += "assign {} = {}[{}];\n".format(getParentWireName(obj, obj.bits[i]), getParentWireName(obj, obj.a), i)
         
     return str
 
+def InlineBitsMSBF(obj:Logic):
+    str = ""
+    w = len(obj.bits)
+    if (w == 1):
+        return "assign {} = {};\n".format(getParentWireName(obj, obj.bits[0]), getParentWireName(obj, obj.a))
+
+    for i in range(w):
+        str += "assign {} = {}[{}];\n".format(getParentWireName(obj, obj.bits[i]), getParentWireName(obj, obj.a), i)
+        
+    return str
+
+def InlineRepeat(obj:Logic):
+    str = ""
+    w = obj.r.getWidth()
+    if (w == 1):
+        return "assign {} = {};\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.i))
+
+    str = 'assign {} ='.format(getParentWireName(obj, obj.r))
+    
+    link = '{'
+    for i in range(w):
+        str += link + "{}".format(getParentWireName(obj, obj.i))
+        link = ','
+    str += '};\n'
+    
+    return str
+
+def InlineConcatenateMSBF(obj:Logic):
+    str = ""
+    w = len(obj.ins)
+    if (w == 1):
+        return "assign {} = {};\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.ins[0]))
+
+    str = 'assign {} ='.format(getParentWireName(obj, obj.r))
+    
+    link = '{'
+    for i in range(w):
+        str += link + "{}".format(getParentWireName(obj, obj.ins[i]))
+        link = ','
+    str += '};\n'
+    
+    return str
+
 def InlineAnd(obj:Logic):
     str = "assign {} =".format(getParentWireName(obj, obj.r))
     link = ""
@@ -262,15 +323,40 @@ def InlineEqual(obj:Logic):
     return "assign {} = ({} == {})? 1:0;\n".format(getParentWireName(obj, obj.r), getParentWireName(obj, obj.a), getParentWireName(obj, obj.b))
 
 def BodyReg(obj:Logic):
+    clkname = getObjectClockDriver(obj).name
     str = "reg "+getWidthInfo(obj.q) + " rq = 0;\n"
-    str += "always @(posedge clk)\n"
-    str +=  "if (e == 1)\n"
-    str += "begin\n"
+    str += "always @(posedge {})\n".format(clkname)
+    close = ""
+    if not(obj.r is None):
+        str += "if (r == 1)\n"
+        str += "begin\n"
+        str += "   rq <= 0;\n"
+        str += "end\n";
+        str += "else\n";
+        str += "begin\n";
+        close = "end\n";
+        
+    if not(obj.e is None):
+        str += "if (e == 1)\n"
+        str += "begin\n"
+        close = "end\n" + close
+        
     str += "   rq <= d;\n"
-    str += "end\n"
+    
+    str += close
     str += "assign q = rq;\n"
     return str
 
+def BodyGatedClock(obj:Logic):
+    str = "reg eq = 0;\n"
+    str += "always @(negedge clk_in)\n"
+    str += "begin\n"
+    str += "   eq <= enin;\n"
+    str += "end\n"
+    str += "assign enout = eq;\n"
+    str += "assign clk_out = enout & clk_in;\n"
+    return str
+
 class VerilogGenerator:
     
     
@@ -284,25 +370,32 @@ class VerilogGenerator:
         self.inlinablePrimitives[And2] = InlineAnd2
         self.inlinablePrimitives[And] = InlineAnd
         self.inlinablePrimitives[Buf] = InlineBuf
-        self.inlinablePrimitives[ZeroExtend] = InlineZeroExtend
+        self.inlinablePrimitives[Bit] = InlineBit
+        self.inlinablePrimitives[BitsLSBF] = InlineBitsLSBF
+        self.inlinablePrimitives[BitsMSBF] = InlineBitsMSBF
+        self.inlinablePrimitives[ConcatenateMSBF] = InlineConcatenateMSBF
         self.inlinablePrimitives[Constant] = InlineConstant
         self.inlinablePrimitives[Equal] = InlineEqual
         self.inlinablePrimitives[EqualConstant] = InlineEqualConstant
+        self.inlinablePrimitives[Nand2] = InlineNand2
         self.inlinablePrimitives[Not] = InlineNot
         self.inlinablePrimitives[Nor] = InlineNor
-        self.inlinablePrimitives[Nand2] = InlineNand2
-        self.inlinablePrimitives[Or2] = InlineOr2
+        self.inlinablePrimitives[Nor2] = InlineNor2
         self.inlinablePrimitives[Or] = InlineOr
+        self.inlinablePrimitives[Or2] = InlineOr2
         self.inlinablePrimitives[Mux2] = InlineMux2
         self.inlinablePrimitives[Sub] = InlineSub
-        self.inlinablePrimitives[Bits] = InlineBits
         self.inlinablePrimitives[Xor2] = InlineXor2
+        self.inlinablePrimitives[Range] = InlineRange
+        self.inlinablePrimitives[Repeat] = InlineRepeat
+        self.inlinablePrimitives[ZeroExtend] = InlineZeroExtend
         
         self.providingBody = {}
         
         self.providingBody[Reg] = BodyReg 
+        self.providingBody[GatedClock] = BodyGatedClock
         
-    def getVerilogForHierarchy(self, obj=None):
+    def getVerilogForHierarchy(self, obj=None, noInstanceNumberInTopEntity=False):
         """
         Generates Verilog for all entities of the object hierarchy
 
@@ -314,24 +407,24 @@ class VerilogGenerator:
         if (obj is None):
             obj = self.obj
             
-        str = self.getVerilog(obj)
+        str = self.getVerilog(obj, noInstanceNumber = noInstanceNumberInTopEntity)
         
-        for child in obj.children:
+        for child in obj.children.values():
             if (not(self.isInlinable(child))):
                 # skip inlinable modules from verilog generation
                 str += "\n"
-                str += self.getVerilogForHierarchy(child)
+                str += self.getVerilogForHierarchy(child, noInstanceNumberInTopEntity=False)
 
         return str        
         
-    def getVerilog(self, obj=None):
+    def getVerilog(self, obj=None, noInstanceNumber=False):
         str = "// This file was automatically created by py4hw RTL generator\n"
         
         if (obj is None):
             obj = self.obj
             
             
-        str += self.createModuleHeader(obj)
+        str += self.createModuleHeader(obj, noInstanceNumber=noInstanceNumber)
         
         localWires = collectLocalWires(obj)
         wireNames = getWireNames(obj)
@@ -346,14 +439,19 @@ class VerilogGenerator:
                 
         if (obj.isPropagatable()):
             # generate code from propagate function
-            pass
+            if (self.isProvidingBody(obj)):
+                str += self.provideBody(obj);
+            else:
+                print('transpiling', obj.getFullPath())
+                str += self.generateCodeFromPropagate(obj)
+
         elif (obj.isClockable()):
             #generate code from clock function
             if (self.isProvidingBody(obj)):
                 str += self.provideBody(obj);
             else:
                 str += self.generateCodeFromClock(obj)
-            pass
+
         else:
             # structural circuit
             str += self.createModuleInstances(obj)
@@ -379,23 +477,30 @@ class VerilogGenerator:
         if (obj.isPrimitive()):
             return obj.isClockable()
         
-        for child in obj.children:
+        for child in obj.children.values():
             if (self.anyClockableDescendant(child)):
                 return True
             
         return False
         
-    def createModuleHeader(self, obj:Logic):
-        str = "module " + getVerilogModuleName(obj) + " (\n\t"
+    def createModuleHeader(self, obj:Logic, noInstanceNumber=None):
+        str = "module " + getVerilogModuleName(obj, noInstanceNumber=noInstanceNumber) + " (\n\t"
+
+        link = ""
         
+        if (isinstance(obj, GatedClock)):
+            drv = obj.drv
+            str += link + "input clk_in".format(drv.base.name)
+            link = ",\n\t"
+            str += link + "output clk_out".format(drv.name)
+            
         reg = ""
         if (obj.isClockable() and not(self.isProvidingBody(obj))):
             reg = " reg "
         
-        link = ""
-        
         if (self.anyClockableDescendant(obj)):
-            str += "input clk"
+            clkname = getObjectClockDriver(obj).name
+            str += "input {}".format(clkname)
             link = ",\n\t"
         
         for inp in obj.inPorts:
@@ -446,7 +551,7 @@ class VerilogGenerator:
     def createModuleInstances(self, obj:Logic):
         str = "\n"
         
-        for child in obj.children:
+        for child in obj.children.values():
             
             if (self.isInlinable(child)):
                 str += self.inlinePrimitive(child)
@@ -460,11 +565,20 @@ class VerilogGenerator:
         str += "("
         link = ""
         
-        # Clock is an implicit parameter
-        if (self.anyClockableDescendant(child)):
-            str += link + ".clk(clk)"
+        if (isinstance(child, GatedClock)):
+            # ClockGate elements are special
+            drv:ClockDriver = child.drv
+            str += link + ".clk_in({})".format(drv.base.name)
+            link = ","
+            str += link + ".clk_out({})".format(drv.name)
+            
+        elif (self.anyClockableDescendant(child)):
+            # Clock is an implicit parameter
+            clkname = getObjectClockDriver(child).name
+            str += link + ".{}({})".format(clkname, clkname)
             link = ","
         
+
         # get the wire names from the instantiator
         wireName = getWireNames(child.parent)
         
@@ -485,6 +599,7 @@ class VerilogGenerator:
     
     def generateCodeFromClock(self, obj:Logic):
         str = "// Code generated from clock method\n"
+        clkname = getObjectClockDriver(obj).name
         
         tr = Python2VerilogTranspiler(obj, 'clock')
         
@@ -493,7 +608,7 @@ class VerilogGenerator:
         str += "// local declarations\n"
         str += tr.getExtraDeclarations() + "\n";
         str += "// sequential process\n"
-        str += "always @(posedge clk)\n"
+        str += "always @(posedge {})\n".format(clkname)
         str += "begin\n"
         str += transpiled + "\n"
         str += "end\n"
@@ -502,16 +617,68 @@ class VerilogGenerator:
         
         return str
 
+    def generateCodeFromPropagate(self, obj:Logic):
+        str = "// Code generated from clock method\n"
+        clkname = getObjectClockDriver(obj).name
+        
+        tr = Python2VerilogTranspiler(obj, 'propagate')
+        
+        transpiled = tr.transpile() ;
+        
+        str += "// local declarations\n"
+        str += tr.getExtraDeclarations() + "\n";
+        str += "// sequential process\n"
+        str += "always @(posedge {})\n".format(clkname)
+        str += "begin\n"
+        str += transpiled + "\n"
+        str += "end\n"
+        
+        str += "\n"
+        
+        return str
 import inspect
 import ast
 import textwrap
     
+
+def getAstValue(obj):
+    if (isinstance(obj, ast.Constant)):
+        return getAstValue(obj.value)
+    elif (isinstance(obj, str)):
+        return obj
+    elif (isinstance(obj, int)):
+        return obj
+    if (isinstance(obj, ast.Name)):
+        return obj.id
+    else:
+        print('DECODING VALUE:', type(obj))
+        return obj
+    
+def getAstName(obj):
+    if (isinstance(obj, ast.Name)):
+        return obj.id
+    elif (isinstance(obj, ast.Attribute)):
+        return getAstName(obj.attr)
+    elif (isinstance(obj, str)):
+        return obj
+    elif (isinstance(obj, list)):
+        if (len(obj)>1):
+            raise Exception('multiple names in {}'.format(obj))
+        
+        return getAstName(obj[0])
+    else:
+        raise Exception('unknown type {}'.format(type(obj)))
+
 class Python2VerilogTranspiler:
 
     def __init__(self, obj:Logic, methodName:str):
         self.obj = obj
         self.methodName = methodName
         self.signals = {}
+        self.indent = 0
+
+    def getIndent(self):
+        return ' ' * (self.indent * 4)
         
     def transpile(self):
         
@@ -546,8 +713,13 @@ class Python2VerilogTranspiler:
 
         extra = [x for x in self.signals if x not in portNames]
         
+        # @todo we should analyze the number of possible values of 
+        # extra signals to decide their width. By now we consider a worst
+        # case scenario with extra signals all requiring a maximum of 8 bits
+        # this will be simplyfied during synthesis if less bits are required
+        # but it will cause a BUG if the required bits are higher
         for sig in extra:
-            str += "reg " + sig + " = 0;" 
+            str += "reg [7:0] " + sig + " = 0;\n" 
             
         return str
     
@@ -565,6 +737,8 @@ class Python2VerilogTranspiler:
             return self.transpileName(line)
         if (type(line) == ast.Eq):
             return "==";
+        if (type(line) == ast.Constant):
+            return "{}".format(getAstValue(line))
         if (type(line) == ast.Num):
             return "{}".format(int(line.n))
         
@@ -595,25 +769,28 @@ class Python2VerilogTranspiler:
         if (len(targets) > 1):
             return ast.dump(line)
         
-        target:ast.Attribute = targets[0]
-        
-        var = target.attr
+        var = getAstName(targets[0])
         
         self.signals[var] = var
         
-        return var + " <= " + self.transpileUnknown(line.value) + ";\n";
+        return var + " <= " + self.transpileUnknown(line.value) + ";\n" + self.getIndent()  ;
     
     def transpileIf(self, line:ast.If):
         str = "if " + self.transpileUnknown(line.test) + "\n"
-        str += "begin\n"
-        str += self.transpileUnknown(line.body)
-        str += "end\n"
+        str += self.getIndent() + "begin\n"
+        self.indent += 1
+        str += self.getIndent() + self.transpileUnknown(line.body)
+        self.indent -= 1
+        str +=  "end\n"
         
         if (not(line.orelse is None)):
-            str += "else \n"
-            str += "begin\n"
-            str += self.transpileUnknown(line.orelse)
-            str += "end\n"
+            str += self.getIndent() + "else \n"
+            str += self.getIndent() + "begin\n"
+            self.indent += 1
+            str += self.getIndent() + self.transpileUnknown(line.orelse)
+            self.indent -= 1
+            str +=  "end\n"
+
         return str
     
     def transpileCompare(self, line:ast.Compare):
@@ -631,21 +808,23 @@ class Python2VerilogTranspiler:
     
     
     
+
 class ReplaceWireGets(ast.NodeTransformer):
     def visit_Call(self, node):
-        attr = node.func.attr
+        attr = getAstName(node.func)
         
         if (attr == 'get'):
-            if isinstance(node.func.value, ast.Attribute):
-                wirename = node.func.value.attr
-                #print('REPLACING GET ', wirename)
-                return ast.Name(wirename, ast.Load)
+            #if isinstance(node.func.value, ast.Attribute):
+            wirename = getAstName(node.func.value)
+            #print('REPLACING GET ', wirename)
+            return ast.Name(wirename, ast.Load)
         
         return node
     
+    
 class ReplaceWirePrepare(ast.NodeTransformer):
     def visit_Call(self, node):
-        attr = node.func.attr
+        attr = getAstName(node.func)
         
         if (attr == 'prepare'):
             #print('REPLACE WIRE PUTS FUNC:', attr , node.func.value.attr, node.args)
